/**
 * @file    Camera_android.fsh
 * @brief   Camera
 * @since   2015/03/24
 *
 * Copyright (c)2015 Nintendo Co., Ltd. All rights reserved.
 */

precision mediump float;

// ----------------------------------------
// フラグメントシェーダーに入力される uniform 変数
uniform sampler2D uYTexture; //!< 入力: テクスチャー
uniform sampler2D uUVTexture; //!< 入力: テクスチャー

// ----------------------------------------
// フラグメントシェーダーに入力される varying 変数
varying mediump vec2 vTexcoord0; //!< 出力: テクスチャー座標

// ------------------------------------------------------------
// フラグメントシェーダーのエントリーポイント
// ------------------------------------------------------------
void main()
{
	vec4 textureColor = vec4(0.0, 0.0, 0.0, 1.0);
	
	float yValue = texture2D(uYTexture, vTexcoord0.xy).r;
	vec4 uvValue = texture2D(uUVTexture, vTexcoord0.xy);

	float Y = 255.0 * yValue;
	float U = 255.0 * uvValue.a;
	float V = 255.0 * uvValue.r;

	if (Y < 16.0)
		Y = 0.0;
	else
		Y = Y - 16.0;
	
	Y = Y / 239.0;
	U = (U - 128.0) / 128.0 * 0.436;
	V = (V - 128.0) / 128.0 * 0.615;

	textureColor.r = clamp(Y + 1.13983 * V, 0.0, 1.0);
	textureColor.g = clamp(Y - 0.39465 * U - 0.5806 * V, 0.0, 1.0);
	textureColor.b = clamp(Y + 2.03211 * U, 0.0, 1.0);

	gl_FragColor = textureColor;
}